/**
 * @author Hamadou BA  
 * Depot GitLab : https://www-apps.univ-lehavre.fr/forge/bh243413/tp2-ps-resolution-de-systemes-lineaires.git  
 *  
 * Classe representant la methode de resolution de systemes lineaires par factorisation LDR
 */
package AlgLin;

public class Helder extends SysLin {
	private SysTriangInfUnite L;
	private SysDiagonal D;
	private SysTriangSupUnite R;

	public Helder(Matrice matrice, Vecteur secondMembre) throws IrregularSysLinException {
		super(matrice, secondMembre);
	}

	public void factorLDR() {
		int degre = getOrdre();
		Matrice matriceL = new Matrice(degre, degre);
		Matrice matriceD = new Matrice(degre, degre);
		Matrice matriceR = new Matrice(degre, degre);
		Matrice matriceA = getMatriceSystem();

		for (int i = 0; i < degre; i++) {
			for (int j = 0; j < degre; j++) {
				if (i == j) {
					matriceL.remplacecoef(i, j, 1.0);
					matriceR.remplacecoef(i, j, 1.0);
				} else {
					matriceL.remplacecoef(i, j, 0.0);
					matriceR.remplacecoef(i, j, 0.0);
				}
			}
		}

		for (int j = 0; j < degre; j++) {
			double diagonalSum = 0.0;
			for (int k = 0; k < j; k++) {
				diagonalSum += Math.pow(matriceL.getCoeff(j, k) * matriceD.getCoeff(k, k) * matriceR.getCoeff(k, j), 2);
			}
			double diagonalValue = matriceA.getCoeff(j, j) - diagonalSum;

			if (Math.abs(diagonalValue) < Matrice.EPSILON) {
				diagonalValue = Matrice.EPSILON;
			}

			matriceD.remplacecoef(j, j, diagonalValue);

			for (int i = j + 1; i < degre; i++) {
				double sumL = 0.0;
				for (int k = 0; k < j; k++) {
					sumL += matriceL.getCoeff(i, k) * matriceD.getCoeff(k, k) * matriceR.getCoeff(k, j);
				}
				double lValue = (matriceA.getCoeff(i, j) - sumL) / matriceD.getCoeff(j, j);
				matriceL.remplacecoef(i, j, lValue);

				double sumR = 0.0;
				for (int k = 0; k < j; k++) {
					sumR += matriceL.getCoeff(j, k) * matriceD.getCoeff(k, k) * matriceR.getCoeff(k, i);
				}
				double rValue = (matriceA.getCoeff(j, i) - sumR) / matriceD.getCoeff(j, j);
				matriceR.remplacecoef(j, i, rValue);
			}
		}

		try {
			L = new SysTriangInfUnite(matriceL, getSecondMembre());
			D = new SysDiagonal(matriceD, getSecondMembre());
			R = new SysTriangSupUnite(matriceR, getSecondMembre());
		} catch (IrregularSysLinException e) {
			System.out.println("Erreur lors de la cré]eation des systè]emes");
		}
	}

	@Override
	public Vecteur resolution() {
		factorLDR();
		Vecteur x = new Vecteur(getOrdre());
		Vecteur y = new Vecteur(getOrdre());
		Vecteur z = new Vecteur(getOrdre());
		try {
			L.setSecondMembre(getSecondMembre());
			y = L.resolution();
			D.setSecondMembre(y);
			z = D.resolution();
			R.setSecondMembre(z);
			x = R.resolution();
		} catch (IrregularSysLinException e) {
			System.out.println("Erreur lors de la resolution");
		}
		return x;
	}

	public Vecteur resolutionPartielle() {
		Vecteur x = new Vecteur(getOrdre());
		Vecteur y = new Vecteur(getOrdre());
		Vecteur z = new Vecteur(getOrdre());
		try {
			L.setSecondMembre(getSecondMembre());
			y = L.resolution();
			D.setSecondMembre(y);
			z = D.resolution();
			R.setSecondMembre(z);
			x = R.resolution();
		} catch (IrregularSysLinException e) {
			System.out.println("Erreur lors de la resolution");
		}
		return x;
	}

	public void setSecondMembre(Vecteur nouveauSecondMembre) {
		super.setSecondMembre(nouveauSecondMembre);
		L.setSecondMembre(nouveauSecondMembre);
		D.setSecondMembre(nouveauSecondMembre);
		R.setSecondMembre(nouveauSecondMembre);
	}

	public SysTriangInfUnite getL() {
		return L;
	}

	public SysDiagonal getD() {
		return D;
	}

	public SysTriangSupUnite getR() {
		return R;
	}

	public static void main(String[] args) throws Exception {
		// Cas 1 : Ax = b
		System.out.println("Cas 1 : Ax = b");
		Matrice matrice = new Matrice("./resources/matriceTest.txt");
		Vecteur secondMembre = new Vecteur("./resources/secondMembreTest.txt");

		Helder helder = new Helder(matrice, secondMembre);
		Vecteur solution = helder.resolution();

		System.out.println("Matrice L :\n" + helder.getL().getMatriceSystem().toString());
		System.out.println("Matrice D :\n" + helder.getD().getMatriceSystem().toString());
		System.out.println("Matrice R :\n" + helder.getR().getMatriceSystem().toString());
		System.out.println("Solution x de Ax = b : " + solution.toString());

		// Cas 2 : A²x = b
		System.out.println("\nCas 2 : A²x = b");
		Vecteur y = solution;
		Helder helder2 = new Helder(matrice, y);
		Vecteur solutionX = helder2.resolution();
		System.out.println("Solution x de A²x = b : " + solutionX.toString());

		// Cas 3 : Ax - b
		System.out.println("\nCas 3 : Ax - b");
		Vecteur Ax = Vecteur.produit(matrice, solution);
		System.out.println("Ax : " + Ax.toString());
		helder.setSecondMembre(new Vecteur("./resources/secondMembreTest2.txt"));
		System.out.println("-b : " + helder.getSecondMembre().toString());

		Vecteur Ax_b = Vecteur.addition(Ax, helder.getSecondMembre());

		// Test de la norme
		double normeL1 = Vecteur.normeL1(Ax_b);
		double normeL2 = Vecteur.normeL2(Ax_b);
		double normeInf = Vecteur.normeLinfini(Ax_b);

		System.out.println("\nNormes du vecteur Ax - b :");
		System.out.println("Norme L1 : " + normeL1);
		System.out.println("Norme L2 : " + normeL2);
		System.out.println("Norme Infinie : " + normeInf);
	}
}